@@ -1,6 +1,6 @@
 /*  test/test-vcf-api.c -- VCF test harness.
 
-    Copyright (C) 2013, 2014, 2017-2019 Genome Research Ltd.
+    Copyright (C) 2013, 2014, 2017-2020 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -24,11 +24,14 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include <config.h>
 
+#include <errno.h>
 #include <stdio.h>
-#include <htslib/hts.h>
-#include <htslib/vcf.h>
-#include <htslib/kstring.h>
-#include <htslib/kseq.h>
+#include <string.h>
+
+#include "../htslib/hts.h"
+#include "../htslib/vcf.h"
+#include "../htslib/kstring.h"
+#include "../htslib/kseq.h"
 
 void error(const char *format, ...)
 {
@@ -43,6 +46,65 @@ void error(const char *format, ...)
 #define STRINGIFY(x) #x
 #define check0(x) ((x) == 0 ? (void) 0 : error("Failed: %s", STRINGIFY(x)))
 
+static int check_alleles(bcf1_t *rec, const char **alleles, int num) {
+    int i;
+    if (rec->n_allele !=  num) {
+        fprintf(stderr, "Wrong number of alleles - expected %d, got %d\n",
+                num, rec->n_allele);
+        return -1;
+    }
+    if (bcf_unpack(rec, BCF_UN_STR) != 0)
+        return -1;
+    for (i = 0; i < num; i++) {
+        if (0 != strcmp(alleles[i], rec->d.allele[i])) {
+            fprintf(stderr,
+                    "Mismatch for allele %d : expected '%s' got '%s'\n",
+                    i, alleles[i], rec->d.allele[i]);
+            return -1;
+        }
+    }
+    return 0;
+}
+
+static void test_update_alleles(bcf_hdr_t *hdr, bcf1_t *rec)
+{
+    // Exercise bcf_update_alleles() a bit
+    const char *alleles1[2] = { "G", "A" };
+    const char *alleles2[3] = { "C", "TGCA", "CATG" };
+#define rep10(x) x x x x x x x x x x
+    const char *alleles3[3] = { rep10("ATTCTAGATC"), "TGCA",
+                                rep10("CTATTATCTCTAATGACATG") };
+#undef rep10
+    const char *alleles4[3] = { alleles3[2], NULL, alleles3[0] };
+    // Add some alleles
+    check0(bcf_update_alleles(hdr, rec, alleles1, 2));
+    check0(check_alleles(rec, alleles1, 2));
+    // Erase them
+    check0(bcf_update_alleles(hdr, rec, NULL, 0));
+    check0(check_alleles(rec, NULL, 0));
+    // Expand to three
+    check0(bcf_update_alleles(hdr, rec, alleles2, 3));
+    check0(check_alleles(rec, alleles2, 3));
+    // Now try some bigger ones (should force a realloc)
+    check0(bcf_update_alleles(hdr, rec, alleles3, 3));
+    check0(check_alleles(rec, alleles3, 3));
+    // Ensure it works even if one of the alleles points into the
+    // existing structure
+    alleles4[1] = rec->d.allele[1];
+    check0(bcf_update_alleles(hdr, rec, alleles4, 3));
+    alleles4[1] = alleles3[1]; // Will have been clobbered by the update
+    check0(check_alleles(rec, alleles4, 3));
+    // Ensure it works when the alleles point into the existing data,
+    // rec->d.allele is used to define the input array and the
+    // order of the entries is changed.  The result of this should
+    // be the same as alleles2.
+    char *tmp = rec->d.allele[0] + strlen(rec->d.allele[0]) - 4;
+    rec->d.allele[0] = rec->d.allele[2] + strlen(rec->d.allele[2]) - 1;
+    rec->d.allele[2] = tmp;
+    check0(bcf_update_alleles(hdr, rec, (const char **) rec->d.allele, 3));
+    check0(check_alleles(rec, alleles2, 3));
+}
+
 void write_bcf(char *fname)
 {
     // Init
@@ -53,6 +115,9 @@ void write_bcf(char *fname)
     bcf1_t *rec    = bcf_init1();
     if (!rec) error("bcf_init1 : %s", strerror(errno));
 
+    // Check no-op on fresh bcf1_t
+    check0(bcf_update_alleles(hdr, rec, NULL, 0));
+
     // Create VCF header
     kstring_t str = {0,0,0};
     check0(bcf_hdr_append(hdr, "##fileDate=20090805"));
@@ -108,7 +173,10 @@ void write_bcf(char *fname)
     // .. ID
     check0(bcf_update_id(hdr, rec, "rs6054257"));
     // .. REF and ALT
+    test_update_alleles(hdr, rec);
+    const char *alleles[2] = { "G", "A" };
     check0(bcf_update_alleles_str(hdr, rec, "G,A"));
+    check0(check_alleles(rec, alleles, 2));
     // .. QUAL
     rec->qual = 29;
     // .. FILTER
@@ -463,6 +531,102 @@ void test_get_format_values(const char *fname)
     check_format_values(fname);
 }
 
+void test_invalid_end_tag(void)
+{
+    static const char vcf_data[] = "data:,"
+        "##fileformat=VCFv4.1\n"
+        "##contig=<ID=X,length=155270560>\n"
+        "##INFO=<ID=END,Number=1,Type=Integer,Description=\"End coordinate of this variant\">\n"
+        "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n"
+        "X\t86470037\trs59780433a\tTTTCA\tTGGTT,T\t.\t.\tEND=85725113\n"
+        "X\t86470038\trs59780433b\tT\tTGGTT,T\t.\t.\tEND=86470047\n";
+
+    htsFile *fp;
+    bcf_hdr_t *hdr;
+    bcf1_t *rec;
+    int ret;
+    int32_t tmpi;
+    enum htsLogLevel logging = hts_get_log_level();
+
+    // Silence warning messages
+    hts_set_log_level(HTS_LOG_ERROR);
+
+    fp = hts_open(vcf_data, "r");
+    if (!fp) error("Failed to open vcf data : %s", strerror(errno));
+    rec = bcf_init1();
+    if (!rec) error("Failed to allocate BCF record : %s", strerror(errno));
+
+    hdr = bcf_hdr_read(fp);
+    if (!hdr) error("Failed to read BCF header : %s", strerror(errno));
+
+    check0(bcf_read(fp, hdr, rec));
+    // rec->rlen should ignore the bogus END tag value on the first read
+    if (rec->rlen != 5) {
+        error("Incorrect rlen - expected 5 got %"PRIhts_pos"\n", rec->rlen);
+    }
+
+    check0(bcf_read(fp, hdr, rec));
+    // While on the second it should use it
+    if (rec->rlen != 10) {
+        error("Incorrect rlen - expected 10 got %"PRIhts_pos"\n", rec->rlen);
+    }
+
+    // Try to break it - will change rlen
+    tmpi = 85725113;
+    check0(bcf_update_info_int32(hdr, rec, "END", &tmpi, 1));
+
+    if (rec->rlen != 1) {
+        error("Incorrect rlen - expected 1 got %"PRIhts_pos"\n", rec->rlen);
+    }
+
+    ret = bcf_read(fp, hdr, rec);
+    if (ret != -1) {
+        error("Unexpected return code %d from bcf_read at EOF", ret);
+    }
+
+    bcf_destroy1(rec);
+    bcf_hdr_destroy(hdr);
+    ret = hts_close(fp);
+    if (ret != 0) {
+        error("Unexpected return code %d from hts_close", ret);
+    }
+
+    hts_set_log_level(logging);
+}
+
+void test_open_format() {
+    char mode[5];
+    int ret;
+    strcpy(mode, "r");
+    ret = vcf_open_mode(mode+1, "mode1.bcf", NULL);
+    if (strncmp(mode, "rb", 2) || ret)
+        error("Mode '%s' does not match the expected value '%s'", mode, "rb");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.vcf", NULL);
+    if (strncmp(mode, "r", 1) || ret)
+        error("Mode '%s' does not match the expected value '%s'", mode, "r");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.vcf.gz", NULL);
+    if (strncmp(mode, "rz", 2) || ret)
+        error("Mode '%s' does not match the expected value '%s'", mode, "rz");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.vcf.bgz", NULL);
+    if (strncmp(mode, "rz", 2) || ret)
+        error("Mode '%s' does not match the expected value '%s'", mode, "rz");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.xcf", NULL);
+    if (!ret)
+        error("Expected failure for wrong extension 'xcf'");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.vcf.gbz", NULL);
+    if (!ret)
+        error("Expected failure for wrong extension 'vcf.gbz'");
+    mode[1] = 0;
+    ret = vcf_open_mode(mode+1, "mode1.bvcf.bgz", NULL);
+    if (!ret)
+        error("Expected failure for wrong extension 'vcf.bvcf.bgz'");
+}
+
 int main(int argc, char **argv)
 {
     char *fname = argc>1 ? argv[1] : "rmme.bcf";
@@ -477,6 +641,7 @@ int main(int argc, char **argv)
 
     // additional tests. quiet unless there's a failure.
     test_get_info_values(fname);
-
+    test_invalid_end_tag();
+    test_open_format();
     return 0;
 }
